"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

preprocess_locations.py
~~~~~~~~~~~~~~~~~~~~~~~

Functions to deal with locations and their configuration.

"""

import math

from calliope.exceptions import ModelError, warn
from calliope.core.attrdict import AttrDict
from calliope.preprocess.util import vincenty
from calliope.preprocess.checks import DEFAULTS, POSSIBLE_COSTS


def process_locations(model_config, modelrun_techs):
    """
    Process locations by taking an AttrDict that may include compact keys
    such as ``1,2,3``, and returning an AttrDict with:

    * exactly one key per location with all of its settings
    * fully resolved installed technologies for each location
    * fully expanded transmission links for each location

    Parameters
    ----------
    model_config : AttrDict
    modelrun_techs : AttrDict

    Returns
    -------
    locations : AttrDict
    locations_comments : AttrDict

    """
    techs_in = model_config.techs.copy()
    tech_groups_in = model_config.tech_groups
    locations_in = model_config.locations
    links_in = model_config.get("links", AttrDict())

    allowed_from_file = DEFAULTS.model.file_allowed

    warnings = []
    errors = []
    locations_comments = AttrDict()

    ##
    # Expand compressed `loc1,loc2,loc3,loc4: ...` definitions
    ##
    locations = AttrDict()
    for key in locations_in:
        if ("--" in key) or ("," in key):
            key_locs = explode_locations(key)
            for subkey in key_locs:
                _set_loc_key(locations, subkey, locations_in[key])
        else:
            _set_loc_key(locations, key, locations_in[key])

    ##
    # Kill any locations that the modeller does not want to exist
    ##
    for loc in list(locations.keys()):
        if not locations[loc].get("exists", True):
            locations.del_key(loc)

    ##
    # Process technologies
    ##
    techs_to_delete = []
    for tech_name in techs_in:
        if not techs_in[tech_name].get("exists", True):
            techs_to_delete.append(tech_name)
            continue
        # Get inheritance chain generated in process_techs()
        inheritance_chain = modelrun_techs[tech_name].inheritance

        # Get and save list of required_constraints from base technology
        base_tech = inheritance_chain[-1]
        rq = model_config.tech_groups[base_tech].required_constraints
        # locations[loc_name].techs[tech_name].required_constraints = rq
        techs_in[tech_name].required_constraints = rq

    # Kill any techs that the modeller does not want to exist
    for tech_name in techs_to_delete:
        del techs_in[tech_name]

    ##
    # Fully expand all installed technologies for the location,
    # filling in any undefined parameters from defaults
    ##
    location_techs_to_delete = []

    for loc_name, loc in locations.items():

        if "techs" not in loc:
            # Mark this as a transmission-only node if it has not allowed
            # any technologies
            locations[loc_name].transmission_node = True
            locations_comments.set_key(
                "{}.transmission_node".format(loc_name),
                "Automatically inserted: specifies that this node is "
                "a transmission-only node.",
            )
            continue  # No need to process any technologies at this node

        for tech_name in loc.techs:
            if tech_name in techs_to_delete:
                # Techs that were removed need not be further considered
                continue

            if not isinstance(locations[loc_name].techs[tech_name], dict):
                locations[loc_name].techs[tech_name] = AttrDict()

            # Starting at top of the inheritance chain, for each level,
            # check if the level has location-specific group settings
            # and keep merging together the settings, overwriting as we
            # go along.
            tech_settings = AttrDict()
            for parent in reversed(modelrun_techs[tech_name].inheritance):
                # Does the parent group have model-wide settings?
                tech_settings.union(tech_groups_in[parent], allow_override=True)
                # Does the parent group have location-specific settings?
                if (
                    "tech_groups" in locations[loc_name]
                    and parent in locations[loc_name].tech_groups
                ):
                    tech_settings.union(
                        locations[loc_name].tech_groups[parent], allow_override=True
                    )
            # Now overwrite with the tech's own model-wide
            # and location-specific settings
            tech_settings.union(techs_in[tech_name], allow_override=True)
            if tech_name in locations[loc_name].techs:
                tech_settings.union(
                    locations[loc_name].techs[tech_name], allow_override=True
                )

            tech_settings = cleanup_undesired_keys(tech_settings)

            # Resolve columns in filename if necessary
            file_or_df_configs = [
                i
                for i in tech_settings.keys_nested()
                if (
                    isinstance(tech_settings.get_key(i), str)
                    and (
                        "file=" in tech_settings.get_key(i)
                        or "df=" in tech_settings.get_key(i)
                    )
                )
            ]
            for config_key in file_or_df_configs:
                config_value = tech_settings.get_key(config_key, "")
                if ":" not in config_value:
                    config_value = "{}:{}".format(config_value, loc_name)
                    tech_settings.set_key(config_key, config_value)

            tech_settings = check_costs_and_compute_depreciation_rates(
                tech_name, loc_name, tech_settings, warnings, errors
            )

            # Now merge the tech settings into the location-specific
            # tech dict -- but if a tech specifies ``exists: false``,
            # we kill it at this location
            if not tech_settings.get("exists", True):
                location_techs_to_delete.append(
                    "{}.techs.{}".format(loc_name, tech_name)
                )
            else:
                locations[loc_name].techs[tech_name].union(
                    tech_settings, allow_override=True
                )

    for k in location_techs_to_delete:
        locations.del_key(k)

    # Generate all transmission links
    processed_links = AttrDict()
    for link in links_in:
        loc_from, loc_to = [i.strip() for i in link.split(",")]
        # Skip this link entirely if it has been told not to exist
        if not links_in[link].get("exists", True):
            continue
        # Also skip this link - and warn about it - if it links to a
        # now-inexistant (because removed) location
        if loc_from not in locations.keys() or loc_to not in locations.keys():
            warnings.append(
                "Not building the link {},{} because one or both of its "
                "locations have been removed from the model by setting "
                "``exists: false``".format(loc_from, loc_to)
            )
            continue
        processed_transmission_techs = AttrDict()
        for tech_name in links_in[link].techs:
            # Skip techs that have been told not to exist
            # for this particular link
            if not links_in[link].get_key("techs.{}.exists".format(tech_name), True):
                continue
            if tech_name not in processed_transmission_techs:
                tech_settings = AttrDict()
                # Combine model-wide settings from all parent groups
                for parent in reversed(modelrun_techs[tech_name].inheritance):
                    tech_settings.union(tech_groups_in[parent], allow_override=True)
                # Now overwrite with the tech's own model-wide settings
                tech_settings.union(techs_in[tech_name], allow_override=True)

                # Add link-specific constraint overrides
                if links_in[link].techs[tech_name]:
                    tech_settings.union(
                        links_in[link].techs[tech_name], allow_override=True
                    )

                tech_settings = cleanup_undesired_keys(tech_settings)

                tech_settings = process_per_distance_constraints(
                    tech_name,
                    tech_settings,
                    locations,
                    locations_comments,
                    loc_from,
                    loc_to,
                )
                tech_settings = check_costs_and_compute_depreciation_rates(
                    tech_name, link, tech_settings, warnings, errors
                )
                processed_transmission_techs[tech_name] = tech_settings
            else:
                tech_settings = processed_transmission_techs[tech_name]

            processed_links.set_key(
                "{}.links.{}.techs.{}".format(loc_from, loc_to, tech_name),
                tech_settings.copy(),
            )

            processed_links.set_key(
                "{}.links.{}.techs.{}".format(loc_to, loc_from, tech_name),
                tech_settings.copy(),
            )

            # If this is a one-way link, we set the constraints for energy_prod
            # and energy_con accordingly on both parts of the link
            if tech_settings.get_key("constraints.one_way", False):
                processed_links.set_key(
                    "{}.links.{}.techs.{}.constraints.energy_prod".format(
                        loc_from, loc_to, tech_name
                    ),
                    False,
                )
                processed_links.set_key(
                    "{}.links.{}.techs.{}.constraints.energy_con".format(
                        loc_to, loc_from, tech_name
                    ),
                    False,
                )
    locations.union(processed_links, allow_override=True)

    return locations, locations_comments, list(set(warnings)), list(set(errors))


def explode_locations(k):
    """
    Expands the given key ``k``. ``k``s of the form ``'1--3'`` or
    ``'1,2,3'`` are both expanded into the list ``['1', '2', '3']``.

    Can deal with any combination, e.g. ``'1--3,6,9--11,a'`` results in::

        ['1', '2', '3', '6', '9', '10', '11', 'a']

    Always returns a list, even if ``k`` is just a simple key,
    i.e. ``explode_locations('1')`` returns ``['1']``.

    """
    # Ensure sure we don't pass in other things
    assert isinstance(k, str)
    finalkeys = []
    subkeys = k.split(",")
    for sk in subkeys:
        if "--" in sk:
            begin, end = sk.split("--")
            finalkeys += [str(i).strip() for i in range(int(begin), int(end) + 1)]
        else:
            finalkeys += [sk.strip()]
    if finalkeys == [] or finalkeys == [""]:
        raise KeyError("Empty key")
    return finalkeys


def _set_loc_key(d, k, value):
    """Set key ``k`` in ``d`` to ``value```."""
    if not value:
        return None
    if k in d:
        try:
            d[k].union(value.copy())
        except KeyError as e:
            raise KeyError("Problem at location {}: {}".format(k, str(e)))
    else:
        d[k] = value.copy()


def cleanup_undesired_keys(tech_settings):
    # These keys are removed from the constructed `locations` dict after
    # merging across the inheritance chain, as they are contained in the
    # `techs` dict
    # These are dealt with in process_techs(),
    # we do not want them here
    keys_to_kill = [
        "essentials",
        "allowed_constraints",
        "allowed_costs",
        "required_constraints",
        "allowed_group_constraints",
    ]
    for k in keys_to_kill:
        try:
            del tech_settings[k]
        except KeyError:
            pass

    # We also remove any system-wide constraints here,
    # as they should not be accidentally read from or
    # changed in per-location settings later
    # FIXME: Raise warning that these constraints are deleted?
    system_wide_keys = [
        k for k in tech_settings.constraints.keys() if k.endswith("_systemwide")
    ]
    for k in system_wide_keys:
        del tech_settings.constraints[k]

    return tech_settings


def process_per_distance_constraints(
    tech_name, tech_settings, locations, locations_comments, loc_from, loc_to
):
    # Process distance, if any per_distance constraints exist
    if any("per_distance" in i for i in tech_settings.keys_nested(subkeys_as="list")):
        # If no distance was given, we calculate it from coordinates
        if "distance" not in tech_settings:
            # Simple check - earlier sense-checking already ensures
            # that all locations have either lat/lon or x/y coords
            loc1 = locations[loc_from].coordinates
            loc2 = locations[loc_to].coordinates
            if "lat" in locations[loc_from].coordinates:
                distance = vincenty([loc1.lat, loc1.lon], [loc2.lat, loc2.lon])
            else:
                distance = math.sqrt((loc1.x - loc2.x) ** 2 + (loc1.y - loc2.y) ** 2)

            tech_settings.distance = distance
            locations_comments.set_key(
                "{}.links.{}.techs.{}.distance".format(loc_from, loc_to, tech_name),
                "Distance automatically computed from coordinates",
            )

        # Add per-distance values to their not-per-distance cousins
        # FIXME these are hardcoded for now
        if "energy_eff_per_distance" in tech_settings.constraints:
            distance_energy_eff = (
                tech_settings.constraints.energy_eff_per_distance
                ** tech_settings.distance
            )
            tech_settings.constraints.energy_eff = (
                tech_settings.constraints.get_key("energy_eff", 1.0)
                * distance_energy_eff
            )
            del tech_settings.constraints["energy_eff_per_distance"]
            locations_comments.set_key(
                "{}.links.{}.techs.{}.constraints.energy_eff".format(
                    loc_from, loc_to, tech_name
                ),
                "Includes value computed from energy_eff_per_distance",
            )

        for k in tech_settings.get("costs", AttrDict()).keys_nested(subkeys_as="list"):
            if "energy_cap_per_distance" in k:
                energy_cap_costs_per_distance = (
                    tech_settings.costs.get_key(k) * tech_settings.distance
                )
                tech_settings.costs[k.split(".")[0]].energy_cap = (
                    tech_settings.costs[k.split(".")[0]].get_key("energy_cap", 0)
                    + energy_cap_costs_per_distance
                )
                tech_settings.costs.del_key(k)
                locations_comments.set_key(
                    "{}.links.{}.techs.{}.costs.{}".format(
                        loc_from, loc_to, tech_name, k
                    ),
                    "Includes value computed from energy_cap_per_distance",
                )
            elif "purchase_per_distance" in k:
                purchase_costs_per_distance = (
                    tech_settings.costs.get_key(k) * tech_settings.distance
                )
                tech_settings.costs[k.split(".")[0]].purchase = (
                    tech_settings.costs[k.split(".")[0]].get_key("purchase", 0)
                    + purchase_costs_per_distance
                )
                tech_settings.costs.del_key(k)
                locations_comments.set_key(
                    "{}.links.{}.techs.{}.costs.{}".format(
                        loc_from, loc_to, tech_name, k
                    ),
                    "Includes value computed from purchase_per_distance",
                )

    return tech_settings


def check_costs_and_compute_depreciation_rates(
    tech_id, loc_or_link, tech_config, warnings, errors
):
    cost_classes = list(tech_config.get("costs", {}).keys())
    for cost in cost_classes:

        # Warning if a cost is defined without a cost class, which is probably a mistake
        if cost in POSSIBLE_COSTS:
            warnings.append(
                "`{}` at `{}` defines {} as a cost class. "
                "This is probably an indentation mistake.".format(
                    tech_id, loc_or_link, cost
                )
            )

        # Warning if a cost class is empty
        if not isinstance(tech_config.costs[cost], dict):
            warnings.append(
                "Deleting empty cost class `{}` for technology `{}` at `{}`.".format(
                    cost, tech_id, loc_or_link
                )
            )
            tech_config.costs.del_key(cost)
            # If the cost class is empty and deleted, the rest of the loop
            # is skipped entirely (as it expects the cost class to exist)
            continue

        plant_life = tech_config.constraints.get_key("lifetime", 0)
        interest = tech_config.costs[cost].get_key("interest_rate", None)

        # Only need depreciation rate for technologies that have investment costs
        if any(
            any(j in i for j in ["_area", "_cap", "purchase"])
            for i in tech_config.costs[cost].keys()
        ):
            # MUST define lifetime and interest_rate for these technologies
            if plant_life == 0 or interest is None:
                errors.append(
                    "Must specify constraints.lifetime and costs.{0}.interest_rate "
                    "when specifying fixed `{0}` costs for `{1}`. Set lifetime to 1 "
                    "and interest rate to 0 if you do not want them to have an effect".format(
                        cost, tech_id
                    )
                )
                continue
            # interest rate = 0 -> simple depreciation
            if interest == 0:
                warnings.append(
                    "`{}` interest rate of zero for technology {}, setting "
                    "depreciation rate as 1/lifetime.".format(cost, tech_id)
                )
                dep = 1 / plant_life
            # interest rate > 0 -> annualised depreciation
            else:
                dep = (interest * ((1 + interest) ** plant_life)) / (
                    ((1 + interest) ** plant_life) - 1
                )

            if math.isnan(dep) or dep == 0:
                warnings.append(
                    "No investment {0} cost will be incurred for `{1}` as "
                    "depreciation rate is 0 or NaN. Probably caused by "
                    "inifinte plant life".format(cost, tech_id)
                )
                dep = 0

            tech_config.costs[cost]["depreciation_rate"] = dep
        try:
            tech_config.costs[cost].del_key("interest_rate")
        except KeyError:
            pass
        # If, by deleting 'interest_rate', we end up with an empty dict, delete the cost class
        if len(tech_config.costs[cost].keys()) == 0:
            tech_config.costs.del_key(cost)
    # If, by deleting the cost class, we end up with an empty dict, delete the cost key
    if len(cost_classes) > 0 and len(tech_config.costs.keys()) == 0:
        tech_config.del_key("costs")

    return tech_config



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

io.py
~~~~~

Functions to read and save model results.

"""

import os

import xarray as xr

from calliope._version import __version__
from calliope import exceptions
from calliope.core.util.dataset import split_loc_techs


def read_netcdf(path):
    """Read model_data from NetCDF file"""
    with xr.open_dataset(path) as model_data:
        model_data.load()

    calliope_version = model_data.attrs.get("calliope_version", False)
    if calliope_version:
        if not str(calliope_version) in __version__:
            exceptions.warn(
                "This model data was created with Calliope version {}, "
                "but you are running {}. Proceed with caution!".format(
                    calliope_version, __version__
                )
            )

    # FIXME some checks for consistency
    # use check_dataset from the checks module
    # also check the old checking from 0.5.x

    return model_data


def save_netcdf(model_data, path, model=None):
    encoding = {k: {"zlib": True, "complevel": 4} for k in model_data.data_vars}

    original_model_data_attrs = model_data.attrs
    model_data_attrs = model_data.attrs.copy()

    if model is not None and hasattr(model, "_model_run"):
        # Attach _model_run and _debug_data to _model_data
        model_run_to_save = model._model_run.copy()
        if "timeseries_data" in model_run_to_save:
            del model_run_to_save["timeseries_data"]  # Can't be serialised!
        model_data_attrs["_model_run"] = model_run_to_save.to_yaml()
        model_data_attrs["_debug_data"] = model._debug_data.to_yaml()

    # Convert boolean attrs to ints
    bool_attrs = [k for k, v in model_data_attrs.items() if isinstance(v, bool)]
    for k in bool_attrs:
        model_data_attrs[k] = int(model_data_attrs[k])

    # Convert None attrs to 'None'
    none_attrs = [k for k, v in model_data_attrs.items() if v is None]
    for k in none_attrs:
        model_data_attrs[k] = "None"

    # Convert `object` dtype coords to string
    # FIXME: remove once xarray issue https://github.com/pydata/xarray/issues/2404 is resolved
    for k, v in model_data.coords.items():
        if v.dtype == "O":
            model_data[k] = v.astype("<U{}".format(max([len(i.item()) for i in v])))

    try:
        model_data.attrs = model_data_attrs
        model_data.to_netcdf(path, format="netCDF4", encoding=encoding)
        model_data.close()  # Force-close NetCDF file after writing
    finally:  # Revert model_data.attrs back
        model_data.attrs = original_model_data_attrs


def save_csv(model_data, path, dropna=True):
    """
    If termination condition was not optimal, filters inputs only, and
    warns that results will not be saved.

    """
    os.makedirs(path, exist_ok=False)

    # a MILP model which optimises to within the MIP gap, but does not fully
    # converge on the LP relaxation, may return as 'feasible', not 'optimal'
    if "termination_condition" not in model_data.attrs or model_data.attrs[
        "termination_condition"
    ] in ["optimal", "feasible"]:
        data_vars = model_data.data_vars
    else:
        data_vars = model_data.filter_by_attrs(is_result=0).data_vars
        exceptions.warn(
            "Model termination condition was not optimal, saving inputs only."
        )

    for var in data_vars:
        in_out = "results" if model_data[var].attrs["is_result"] else "inputs"
        out_path = os.path.join(path, "{}_{}.csv".format(in_out, var))
        series = split_loc_techs(model_data[var], return_as="Series")
        if dropna:
            series = series.dropna()
        series.to_csv(out_path, header=True)


def save_lp(model, path):
    if not model.run_config["backend"] == "pyomo":
        raise IOError("Only the pyomo backend can save to LP.")
    if not hasattr(model, "_backend_model"):
        model.run(build_only=True)
    model._backend_model.write(
        path, format="lp", io_options={"symbolic_solver_labels": True}
    )



#!/usr/bin/env python

from pathlib import Path
from setuptools import setup, find_packages

exec(Path("calliope/_version.py").read_text())  # Sets the __version__ variable

requirements = Path("requirements.txt").read_text().strip().split("\n")
long_description = Path("README.md").read_text()


def get_subdirs(path, glob_string):
    return [
        i.relative_to(str(path) + "/") for i in path.glob(glob_string) if i.is_dir()
    ]


def find_calliope_package_data():
    """Returns a list of found directories with package data files"""
    path = Path("./calliope")
    package_data = ["config/*.yaml", "config/*.html", "test/common/*.yaml"]
    for subdir_level in ["*", "*/*", "*/*/*"]:
        for example_dir in get_subdirs(path, "example_models/" + subdir_level):
            package_data.append(str(example_dir) + "/*.csv")
            package_data.append(str(example_dir) + "/*.yaml")
            package_data.append(str(example_dir) + "/*.rst")
    for test_case_dir in get_subdirs(path, "test/common/*"):
        package_data.append(str(test_case_dir) + "/*.csv")
    print(package_data)
    return package_data


setup(
    name="calliope",
    version=__version__,
    author="Calliope contributors listed in AUTHORS",
    author_email="stefan@pfenninger.org",
    description="A multi-scale energy systems modelling framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    url="https://www.callio.pe/",
    download_url="https://github.com/calliope-project/calliope/releases",
    packages=find_packages(),
    package_data={"calliope": find_calliope_package_data()},
    install_requires=requirements,
    entry_points={"console_scripts": ["calliope = calliope.cli:cli"]},
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords=["energy systems", "optimisation", "mathematical programming"],
)



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

exceptions.py
~~~~~~~~~~~~~

Exceptions and Warnings.

"""

import textwrap
import warnings


# Enable simple format when printing ModelWarnings
formatwarning_orig = warnings.formatwarning


def _formatwarning(message, category, filename, lineno, line=None):
    """Formats ModelWarnings as "Warning: message" without extra crud"""
    if category == ModelWarning:
        return "Warning: " + str(message) + "\n"
    else:
        return formatwarning_orig(message, category, filename, lineno, line)


warnings.formatwarning = _formatwarning


class ModelError(Exception):
    """
    ModelErrors should stop execution of the model, e.g. due to a problem
    with the model formulation or input data.

    """

    pass


class BackendError(Exception):
    pass


class ModelWarning(Warning):
    """
    ModelWarnings should be raised for possible model errors, but
    where execution can still continue.

    """

    pass


class BackendWarning(Warning):
    pass


def warn(message, _class=ModelWarning):
    warnings.warn(message, _class)


def print_warnings_and_raise_errors(warnings=None, errors=None):
    """
    Print warnings and raise ModelError from errors.

    Parameters
    ----------
    warnings : list, optional
    errors : list, optional

    """
    if warnings:
        warn(
            "Possible issues found during model processing:\n"
            + textwrap.indent("\n".join(sorted(list(set(warnings)))), " * ")
        )

    if errors:
        raise ModelError(
            "Errors during model processing:\n"
            + textwrap.indent("\n".join(sorted(list(set(errors)))), " * ")
        )

    return None



from calliope import exceptions
import shutil

import pytest
from pytest import approx
import pandas as pd
import numpy as np
import xarray as xr
import calliope
from calliope.test.common.util import check_error_or_warning


class TestModelPreproccesing:
    def test_preprocess_national_scale(self):
        calliope.examples.national_scale()

    def test_preprocess_time_clustering(self):
        calliope.examples.time_clustering()

    def test_preprocess_time_resampling(self):
        calliope.examples.time_resampling()

    def test_preprocess_urban_scale(self):
        calliope.examples.urban_scale()

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_preprocess_milp(self):
        calliope.examples.milp()

    def test_preprocess_operate(self):
        calliope.examples.operate()

    def test_preprocess_time_masking(self):
        calliope.examples.time_masking()


class TestNationalScaleExampleModelSenseChecks:
    def example_tester(self, solver="cbc", solver_io=None):
        override = {
            "model.subset_time": "2005-01-01",
            "run.solver": solver,
        }

        if solver_io:
            override["run.solver_io"] = solver_io

        model = calliope.examples.national_scale(override_dict=override)
        model.run()

        assert model.results.storage_cap.to_pandas()["region1-1::csp"] == approx(
            45129.950
        )
        assert model.results.storage_cap.to_pandas()["region2::battery"] == approx(
            6675.173
        )

        assert model.results.energy_cap.to_pandas()["region1-1::csp"] == approx(
            4626.588
        )
        assert model.results.energy_cap.to_pandas()["region2::battery"] == approx(1000)
        assert model.results.energy_cap.to_pandas()["region1::ccgt"] == approx(30000)

        assert float(model.results.cost.sum()) == approx(38988.7442)

        assert float(
            model.results.systemwide_levelised_cost.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.063543, abs=0.000001)
        assert float(
            model.results.systemwide_capacity_factor.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.2642256, abs=0.000001)

    def test_nationalscale_example_results_cbc(self):
        self.example_tester()

    @pytest.importorskip("gurobipy")
    def test_nationalscale_example_results_gurobi(self):
        self.example_tester(solver="gurobi", solver_io="python")

    def test_nationalscale_example_results_cplex(self):
        if shutil.which("cplex"):
            self.example_tester(solver="cplex")
        else:
            pytest.skip("CPLEX not installed")

    def test_nationalscale_example_results_glpk(self):
        if shutil.which("glpsol"):
            self.example_tester(solver="glpk")
        else:
            pytest.skip("GLPK not installed")

    def test_considers_supply_generation_only_in_total_levelised_cost(self):
        # calculation of expected value:
        # costs = model.get_formatted_array("cost").sum(dim="locs")
        # gen = model.get_formatted_array("carrier_prod").sum(dim=["timesteps", "locs"])
        # lcoe = costs.sum(dim="techs") / gen.sel(techs=["ccgt", "csp"]).sum(dim="techs")
        model = calliope.examples.national_scale()
        model.run()

        assert model.results.total_levelised_cost.item() == approx(0.067005, abs=1e-5)

    def test_fails_gracefully_without_timeseries(self):
        override = {
            "locations.region1.techs.demand_power.constraints.resource": -200,
            "locations.region2.techs.demand_power.constraints.resource": -400,
            "techs.csp.constraints.resource": 100,
        }
        with pytest.raises(calliope.exceptions.ModelError):
            calliope.examples.national_scale(override_dict=override)


class TestNationalScaleExampleModelInfeasibility:
    def example_tester(self):

        model = calliope.examples.national_scale(
            scenario="check_feasibility", override_dict={"run.cyclic_storage": False}
        )

        model.run()

        assert model.results.termination_condition in [
            "infeasible",
            "other",
        ]  # glpk gives 'other' as result

        assert len(model.results.data_vars) == 0
        assert "energy_cap" not in model._model_data.data_vars

    def test_nationalscale_example_results_cbc(self):
        self.example_tester()


class TestNationalScaleExampleModelOperate:
    def example_tester(self):
        with pytest.warns(calliope.exceptions.ModelWarning) as excinfo:
            model = calliope.examples.national_scale(
                override_dict={"model.subset_time": ["2005-01-01", "2005-01-03"]},
                scenario="operate",
            )
            model.run()

        expected_warning = "Resource capacity constraint defined and set to infinity for all supply_plus techs"

        assert check_error_or_warning(excinfo, expected_warning)
        assert all(
            model.results.timesteps
            == pd.date_range("2005-01", "2005-01-03 23:00:00", freq="H")
        )

    def test_nationalscale_example_results_cbc(self):
        self.example_tester()


class TestNationalScaleExampleModelSpores:
    def example_tester(self, solver="cbc", solver_io=None):

        model = calliope.examples.national_scale(
            override_dict={
                "model.subset_time": ["2005-01-01", "2005-01-03"],
                "run.solver": solver,
                "run.solver_io": solver_io,
            },
            scenario="spores",
        )

        model.run(build_only=True)

        # The initial state of the objective cost class scores should be monetary: 1, spores_score: 0
        assert model._backend_model.objective_cost_class["monetary"].value == 1
        assert model._backend_model.objective_cost_class["spores_score"].value == 0

        model.run(force_rerun=True)
        # Expecting three spores + first optimal run
        assert np.allclose(model.results.spores, [0, 1, 2, 3])

        costs = model.results.cost.sum("loc_techs_cost")
        slack_cost = model._backend_model.group_cost_max[
            ("monetary", "systemwide_cost_max")
        ].value

        # First run is the optimal run, everything else is coming up against the slack cost
        assert costs.loc[{"spores": 0, "costs": "monetary"}] * (
            1 + model.run_config["spores_options"]["slack"]
        ) == approx(slack_cost)
        assert all(
            costs.loc[{"spores": slice(1, None), "costs": "monetary"}]
            <= slack_cost * 1.0001
        )

        # In each iteration, the spores_score has to increase
        assert all(costs.diff("spores").loc[{"costs": "spores_score"}] >= 0)

        # The final state of the objective cost class scores should be monetary: 0, spores_score: 1
        assert model._backend_model.objective_cost_class["monetary"].value == 0
        assert model._backend_model.objective_cost_class["spores_score"].value == 1
        return model._model_data

    def test_nationalscale_example_results_cbc(self):
        self.example_tester()

    @pytest.mark.filterwarnings(
        "ignore:(?s).*`gurobi_persistent`.*:calliope.exceptions.ModelWarning"
    )
    @pytest.importorskip("gurobipy")
    def test_nationalscale_example_results_gurobi(self):
        gurobi_data = self.example_tester("gurobi", "python")
        gurobi_persistent_data = self.example_tester("gurobi_persistent", "python")
        assert np.allclose(gurobi_data.energy_cap, gurobi_persistent_data.energy_cap)
        assert np.allclose(gurobi_data.cost, gurobi_persistent_data.cost)

    @pytest.fixture
    def base_model_data(self):
        model = calliope.examples.national_scale(
            override_dict={
                "model.subset_time": ["2005-01-01", "2005-01-03"],
                "run.solver": "cbc",
            },
            scenario="spores",
        )

        model.run()

        return model._model_data

    @pytest.mark.parametrize("init_spore", (0, 1, 2))
    def test_nationalscale_skip_cost_op_spores(self, base_model_data, init_spore):
        spores_model = calliope.Model(
            config=None, model_data=base_model_data.loc[{"spores": [init_spore + 1]}]
        )
        spores_model._model_data.coords["spores"] = [init_spore]

        spores_model.run_config["spores_options"]["skip_cost_op"] = True

        spores_model.run(force_rerun=True)

        assert set(spores_model.results.spores.values) == set(range(init_spore, 4))
        assert base_model_data.loc[{"spores": slice(init_spore + 1, None)}].equals(
            spores_model._model_data.loc[{"spores": slice(init_spore + 1, None)}]
        )

    def test_fail_with_spores_as_input_dim(self, base_model_data):
        spores_model = calliope.Model(
            config=None, model_data=base_model_data.loc[{"spores": [0, 1]}]
        )
        with pytest.raises(exceptions.ModelError) as excinfo:
            spores_model.run(force_rerun=True)
        assert check_error_or_warning(
            excinfo, "Cannot run SPORES with a SPORES dimension in any input"
        )


class TestNationalScaleResampledExampleModelSenseChecks:
    def example_tester(self, solver="cbc", solver_io=None):
        override = {
            "model.subset_time": "2005-01-01",
            "run.solver": solver,
        }

        if solver_io:
            override["run.solver_io"] = solver_io

        model = calliope.examples.time_resampling(override_dict=override)
        model.run()

        assert model.results.storage_cap.to_pandas()["region1-1::csp"] == approx(
            23563.444
        )
        assert model.results.storage_cap.to_pandas()["region2::battery"] == approx(
            6315.78947
        )

        assert model.results.energy_cap.to_pandas()["region1-1::csp"] == approx(
            1440.8377
        )
        assert model.results.energy_cap.to_pandas()["region2::battery"] == approx(1000)
        assert model.results.energy_cap.to_pandas()["region1::ccgt"] == approx(30000)

        assert float(model.results.cost.sum()) == approx(37344.221869)

        assert float(
            model.results.systemwide_levelised_cost.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.063543, abs=0.000001)
        assert float(
            model.results.systemwide_capacity_factor.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.25, abs=0.000001)

    def test_nationalscale_resampled_example_results_cbc(self):
        self.example_tester()

    def test_nationalscale_resampled_example_results_glpk(self):
        if shutil.which("glpsol"):
            self.example_tester(solver="glpk")
        else:
            pytest.skip("GLPK not installed")


class TestNationalScaleClusteredExampleModelSenseChecks:
    def model_runner(
        self,
        solver="cbc",
        solver_io=None,
        how="closest",
        storage_inter_cluster=False,
        cyclic=False,
        storage=True,
    ):
        override = {
            "model.time.function_options": {
                "how": how,
                "storage_inter_cluster": storage_inter_cluster,
            },
            "run.solver": solver,
            "run.cyclic_storage": cyclic,
        }
        if storage is False:
            override.update({"techs.battery.exists": False, "techs.csp.exists": False})

        if solver_io:
            override["run.solver_io"] = solver_io

        model = calliope.examples.time_clustering(override_dict=override)
        model.run()

        return model

    def example_tester_closest(self, solver="cbc", solver_io=None):
        model = self.model_runner(solver=solver, solver_io=solver_io, how="closest")
        # Full 1-hourly model run: 22312488.670967
        assert float(model.results.cost.sum()) == approx(49670627.15297682)

        # Full 1-hourly model run: 0.296973
        assert float(
            model.results.systemwide_levelised_cost.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.137105, abs=0.000001)

        # Full 1-hourly model run: 0.064362
        assert float(
            model.results.systemwide_capacity_factor.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.064501, abs=0.000001)

    def example_tester_mean(self, solver="cbc", solver_io=None):
        model = self.model_runner(solver=solver, solver_io=solver_io, how="mean")
        # Full 1-hourly model run: 22312488.670967
        assert float(model.results.cost.sum()) == approx(22172253.328)

        # Full 1-hourly model run: 0.296973
        assert float(
            model.results.systemwide_levelised_cost.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.127783, abs=0.000001)

        # Full 1-hourly model run: 0.064362
        assert float(
            model.results.systemwide_capacity_factor.loc[dict(carriers="power")]
            .to_pandas()
            .T["battery"]
        ) == approx(0.044458, abs=0.000001)

    def example_tester_storage_inter_cluster(self):
        model = self.model_runner(storage_inter_cluster=True)

        # Full 1-hourly model run: 22312488.670967
        assert float(model.results.cost.sum()) == approx(21825515.304)

        # Full 1-hourly model run: 0.296973
        assert float(
            model.results.systemwide_levelised_cost.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.100760, abs=0.000001)

        # Full 1-hourly model run: 0.064362
        assert float(
            model.results.systemwide_capacity_factor.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.091036, abs=0.000001)

    def test_nationalscale_clustered_example_closest_results_cbc(self):
        self.example_tester_closest()

    def test_nationalscale_clustered_example_closest_results_glpk(self):
        if shutil.which("glpsol"):
            self.example_tester_closest(solver="glpk")
        else:
            pytest.skip("GLPK not installed")

    def test_nationalscale_clustered_example_mean_results_cbc(self):
        self.example_tester_mean()

    @pytest.mark.skip(
        reason="GLPK is useless and delivering different results on different operating systems"
    )
    def test_nationalscale_clustered_example_mean_results_glpk(self):
        if shutil.which("glpsol"):
            self.example_tester_mean(solver="glpk")
        else:
            pytest.skip("GLPK not installed")

    def test_nationalscale_clustered_example_storage_inter_cluster(self):
        self.example_tester_storage_inter_cluster()

    def test_storage_inter_cluster_cyclic(self):
        model = self.model_runner(storage_inter_cluster=True, cyclic=True)
        # Full 1-hourly model run: 22312488.670967
        assert float(model.results.cost.sum()) == approx(18904055.722)

        # Full 1-hourly model run: 0.296973
        assert float(
            model.results.systemwide_levelised_cost.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.122564, abs=0.000001)

        # Full 1-hourly model run: 0.064362
        assert float(
            model.results.systemwide_capacity_factor.loc[
                {"carriers": "power", "techs": "battery"}
            ].item()
        ) == approx(0.075145, abs=0.000001)

    def test_storage_inter_cluster_no_storage(self):
        with pytest.warns(calliope.exceptions.ModelWarning) as excinfo:
            self.model_runner(storage_inter_cluster=True, storage=False)

        expected_warnings = [
            "Tech battery was removed by setting ``exists: False``",
            "Tech csp was removed by setting ``exists: False``",
        ]
        assert check_error_or_warning(excinfo, expected_warnings)


class TestUrbanScaleExampleModelSenseChecks:
    def example_tester(self, resource_unit, solver="cbc", solver_io=None):
        unit_override = {
            "techs.pv.constraints": {
                "resource": "file=pv_resource.csv:{}".format(resource_unit),
                "resource_unit": "energy_{}".format(resource_unit),
            },
            "run.solver": solver,
        }
        override = {"model.subset_time": "2005-07-01", **unit_override}

        if solver_io:
            override["run.solver_io"] = solver_io

        model = calliope.examples.urban_scale(override_dict=override)
        model.run()

        assert model.results.energy_cap.to_pandas()["X1::chp"] == approx(250.090112)

        # GLPK isn't able to get the same answer both times, so we have to account for that here
        if resource_unit == "per_cap" and solver == "glpk":
            heat_pipe_approx = 183.45825
        else:
            heat_pipe_approx = 182.19260

        assert model.results.energy_cap.to_pandas()["X2::heat_pipes:N1"] == approx(
            heat_pipe_approx
        )

        assert model.results.carrier_prod.sum("timesteps").to_pandas()[
            "X3::boiler::heat"
        ] == approx(0.18720)
        assert model.results.resource_area.to_pandas()["X2::pv"] == approx(830.064659)

        assert float(model.results.carrier_export.sum()) == approx(122.7156)

        # GLPK doesn't agree with commercial solvers, so we have to account for that here
        cost_sum = 430.097399 if solver == "glpk" else 430.089188
        assert float(model.results.cost.sum()) == approx(cost_sum)

    def test_urban_example_results_area(self):
        self.example_tester("per_area")

    @pytest.importorskip("gurobipy")
    def test_urban_example_results_area_gurobi(self):
        self.example_tester("per_area", solver="gurobi", solver_io="python")

    def test_urban_example_results_cap(self):
        self.example_tester("per_cap")

    @pytest.importorskip("gurobipy")
    def test_urban_example_results_cap_gurobi(self):
        self.example_tester("per_cap", solver="gurobi", solver_io="python")

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_milp_example_results(self):
        model = calliope.examples.milp(
            override_dict={
                "model.subset_time": "2005-01-01",
                "run.solver_options.mipgap": 0.001,
            }
        )
        model.run()

        assert model.results.energy_cap.to_pandas()["X1::chp"] == 300
        assert model.results.energy_cap.to_pandas()["X2::heat_pipes:N1"] == approx(
            188.363137
        )

        assert model.results.carrier_prod.sum("timesteps").to_pandas()[
            "X1::supply_gas::gas"
        ] == approx(12363.173036)
        assert float(model.results.carrier_export.sum()) == approx(0)

        assert model.results.purchased.to_pandas()["X2::boiler"] == 1
        assert model.results.units.to_pandas()["X1::chp"] == 1

        assert float(model.results.operating_units.sum()) == 24

        assert float(model.results.cost.sum()) == approx(540.780779)

    def test_operate_example_results(self):
        model = calliope.examples.operate(
            override_dict={"model.subset_time": ["2005-07-01", "2005-07-04"]}
        )
        with pytest.warns(calliope.exceptions.ModelWarning) as excinfo:
            model.run()

        expected_warnings = [
            "Energy capacity constraint removed",
            "Resource capacity constraint defined and set to infinity for all supply_plus techs",
            "Storage cannot be cyclic in operate run mode, setting `run.cyclic_storage` to False for this run",
        ]

        assert check_error_or_warning(excinfo, expected_warnings)

        assert all(
            model.results.timesteps
            == pd.date_range("2005-07", "2005-07-04 23:00:00", freq="H")
        )



from calliope._version import __version__

from calliope import examples, exceptions
from calliope.core import AttrDict, Model, read_netcdf
from calliope.core.util.logging import set_log_verbosity


__title__ = "Calliope"
__author__ = "Calliope contributors listed in AUTHORS"
__copyright__ = "Copyright (C) since 2013 Calliope contributors listed in AUTHORS"



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

network.py
~~~~~~~~~~

Transmission/distribution network constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

ORDER = 10  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]
    run_config = backend_model.__calliope_run_config

    if (
        "loc_techs_symmetric_transmission_constraint" in sets
        and run_config["mode"] != "operate"
    ):
        backend_model.symmetric_transmission_constraint = po.Constraint(
            backend_model.loc_techs_symmetric_transmission_constraint,
            rule=symmetric_transmission_constraint_rule,
        )


def symmetric_transmission_constraint_rule(backend_model, loc_tech):
    """
    Constrain e_cap symmetrically for transmission nodes. Transmission techs only.

    .. container:: scrolling-wrapper

        .. math::

            energy_{cap}(loc1::tech:loc2) = energy_{cap}(loc2::tech:loc1)

    """
    lookup = backend_model.__calliope_model_data["data"]["lookup_remotes"]
    loc_tech_remote = lookup[loc_tech]

    return (
        backend_model.energy_cap[loc_tech] == backend_model.energy_cap[loc_tech_remote]
    )



import os

import pytest  # pylint: disable=unused-import

import calliope
from calliope import exceptions
from calliope.test.common.util import check_error_or_warning


def build_model(override_dict, scenario):
    model_path = os.path.join(
        os.path.dirname(__file__), "common", "test_model", "model.yaml"
    )

    return calliope.Model(model_path, override_dict=override_dict, scenario=scenario)


class TestExistsFalse:
    """
    Test removal of techs, locations, links, and transmission techs
    with the ``exists: False`` configuration option.

    """

    def test_tech_exists_false(self):
        overrides = {"techs.test_storage.exists": False}
        with pytest.warns(exceptions.ModelWarning) as excinfo:
            model = build_model(overrides, "simple_storage,one_day,investment_costs")
        model.run()

        # Ensure what should be gone is gone
        assert "test_storage" not in model.results.coords["techs"].values

        # Ensure warnings were raised
        assert check_error_or_warning(
            excinfo,
            "Tech test_storage was removed by setting ``exists: False`` - not checking the consistency of its constraints at location 0.",
        )

    def test_location_exists_false(self):
        overrides = {"locations.1.exists": False}
        with pytest.warns(exceptions.ModelWarning) as excinfo:
            model = build_model(overrides, "simple_storage,one_day,investment_costs")
        model.run()

        # Ensure what should be gone is gone
        assert "1" not in model._model_data.coords["locs"].values

        # Ensure warnings were raised
        assert check_error_or_warning(
            excinfo,
            "Not building the link 0,1 because one or both of its locations have been removed from the model by setting ``exists: false``",
        )

    def test_location_tech_exists_false(self):
        overrides = {"locations.1.techs.test_storage.exists": False}
        model = build_model(overrides, "simple_storage,one_day,investment_costs")
        model.run()

        # Ensure what should be gone is gone
        assert "1::test_storage" not in model._model_data.coords["loc_techs"].values

    def test_link_exists_false(self):
        overrides = {"links.0,1.exists": False}
        model = build_model(overrides, "simple_storage,one_day,investment_costs")
        model.run()

        # Ensure what should be gone is gone
        assert "loc_techs_transmission" not in model._model_data

    def test_link_tech_exists_false(self):
        overrides = {"links.0,1.techs.test_transmission_elec.exists": False}
        model = build_model(overrides, "simple_storage,one_day,investment_costs")
        model.run()

        # Ensure what should be gone is gone
        assert (
            "0::test_transmission_elec:1"
            not in model._model_data.coords["loc_techs_transmission"].values
        )
        assert (
            "0::test_transmission_heat:1"
            in model._model_data.coords["loc_techs_transmission"].values
        )



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

attrdict.py
~~~~~~~~~~~

Implements the AttrDict class (a subclass of regular dict)
used for managing model configuration.

"""

import io
from pathlib import Path
import logging

import numpy as np
import ruamel.yaml as ruamel_yaml

from calliope.core.util.tools import relative_path

logger = logging.getLogger(__name__)


class __Missing(object):
    def __nonzero__(self):
        return False


_MISSING = __Missing()


def _yaml_load(src):
    """Load YAML from a file object or path with useful parser errors"""

    if not isinstance(src, str):
        try:
            src_name = src.name
        except AttributeError:
            src_name = "<yaml stringio>"
        # Force-load file streams as that allows the parser to print
        # much more context when it encounters an error
        src = src.read()
    else:
        src_name = "<yaml string>"
    try:
        result = ruamel_yaml.safe_load(src)
        if not isinstance(result, dict):
            raise ValueError("Could not parse {} as YAML".format(src_name))
        return result
    except ruamel_yaml.YAMLError:
        logger.error("Parser error when reading YAML " "from {}.".format(src_name))
        raise


class AttrDict(dict):
    """
    A subclass of ``dict`` with key access by attributes::

        d = AttrDict({'a': 1, 'b': 2})
        d.a == 1  # True

    Includes a range of additional methods to read and write to YAML,
    and to deal with nested keys.

    """

    __name__ = "AttrDict"

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__

    def __init__(self, source_dict=None):
        super().__init__()

        if source_dict is not None:
            if isinstance(source_dict, dict):
                self.init_from_dict(source_dict)
            else:
                raise ValueError("Must pass a dict to AttrDict")

    def copy(self):
        """Override copy method so that it returns an AttrDict"""
        return AttrDict(self.as_dict().copy())

    def init_from_dict(self, d):
        """
        Initialize a new AttrDict from the given dict. Handles any
        nested dicts by turning them into AttrDicts too::

            d = AttrDict({'a': 1, 'b': {'x': 1, 'y': 2}})
            d.b.x == 1  # True

        """
        for k, v in d.items():
            # First, keys must be strings, not ints
            if isinstance(k, int):
                k = str(k)

            # Now, assign to the key, handling nested AttrDicts properly
            if isinstance(v, dict):
                self.set_key(k, AttrDict(v))
            elif isinstance(v, list):
                # Modifying the list in-place so that if it is a modified
                # list subclass, e.g. CommentedSeq, it is not killed
                for i in range(len(v)):
                    if isinstance(v[i], dict):
                        v[i] = AttrDict(v[i])
                self.set_key(k, v)
            else:
                self.set_key(k, v)

    @classmethod
    def _resolve_imports(cls, loaded, resolve_imports, base_path=None):
        if (
            isinstance(resolve_imports, bool)
            and resolve_imports is True
            and "import" in loaded
        ):
            loaded_dict = loaded
        elif (
            isinstance(resolve_imports, str)
            and resolve_imports + ".import" in loaded.keys_nested()
        ):
            loaded_dict = loaded.get_key(resolve_imports)
        else:  # Return right away if no importing to be done
            return loaded

        # If we end up here, we have something to import
        imports = loaded_dict.get_key("import")
        if not isinstance(imports, list):
            raise ValueError("`import` must be a list.")

        for k in imports:
            if base_path:
                path = relative_path(base_path, k)
            else:
                path = k
            imported = cls.from_yaml(path)
            # loaded is added to imported (i.e. it takes precedence)
            imported.union(loaded_dict)
            loaded_dict = imported
        # 'import' key itself is no longer needed
        loaded_dict.del_key("import")

        if isinstance(resolve_imports, str):
            loaded.set_key(resolve_imports, loaded_dict)
        else:
            loaded = loaded_dict

        return loaded

    @classmethod
    def from_yaml(cls, f, resolve_imports=True):
        """
        Returns an AttrDict initialized from the given path or
        file object ``f``, which must point to a YAML file. The path can
        be a string or a pathlib.Path.

        Parameters
        ----------

        f : str or pathlib.Path
        resolve_imports : bool or str, optional
            If ``resolve_imports`` is True, top-level ``import:`` statements
            are resolved recursively.
            If ``resolve_imports is False, top-level ``import:`` statements
            are treated like any other key and not further processed.
            If ``resolve_imports`` is a string, such as ``foobar``, import
            statements underneath that key are resolved, i.e. ``foobar.import:``.
            When resolving import statements, anything defined locally
            overrides definitions in the imported file.

        """
        if isinstance(f, str) or isinstance(f, Path):
            with open(
                f,
                "r",
                encoding="utf-8",
            ) as src:
                loaded = cls(_yaml_load(src))
        else:
            loaded = cls(_yaml_load(f))
        loaded = cls._resolve_imports(loaded, resolve_imports, base_path=f)
        return loaded

    @classmethod
    def from_yaml_string(cls, string, resolve_imports=True):
        """
        Returns an AttrDict initialized from the given string, which
        must be valid YAML.

        """
        loaded = cls(_yaml_load(string))
        loaded = cls._resolve_imports(loaded, resolve_imports)
        return loaded

    def set_key(self, key, value):
        """
        Set the given ``key`` to the given ``value``. Handles nested
        keys, e.g.::

            d = AttrDict()
            d.set_key('foo.bar', 1)
            d.foo.bar == 1  # True

        """
        if isinstance(value, dict) and not isinstance(value, AttrDict):
            value = AttrDict(value)
        if "." in key:
            key, remainder = key.split(".", 1)
            try:
                self[key].set_key(remainder, value)
            except KeyError:
                self[key] = AttrDict()
                self[key].set_key(remainder, value)
            except AttributeError:
                if self[key] is None:  # If the value is None, we replace it
                    self[key] = AttrDict()
                    self[key].set_key(remainder, value)
                # Else there is probably something there, and we don't just
                # want to overwrite so stop and warn the user
                else:
                    raise KeyError("Cannot set nested key on non-dict key.")
        else:
            if key in self and isinstance(value, AttrDict):
                for k, v in value.items():
                    self[key].set_key(k, v)
            else:
                self[key] = value

    def get_key(self, key, default=_MISSING):
        """
        Looks up the given ``key``. Like set_key(), deals with nested
        keys.

        If default is anything but ``_MISSING``, the given default is
        returned if the key does not exist.

        """
        if "." in key:
            # Nested key of form "foo.bar"
            key, remainder = key.split(".", 1)
            if default != _MISSING:
                try:
                    value = self[key].get_key(remainder, default)
                except KeyError:
                    # subdict exists, but doesn't contain key
                    return default
                except AttributeError:
                    # key points to non-dict thing, so no get_key attribute
                    return default
            else:
                value = self[key].get_key(remainder)
        else:
            # Single, non-nested key of form "foo"
            if default != _MISSING:
                return self.get(key, default)
            else:
                return self[key]
        return value

    def del_key(self, key):
        """Delete the given key. Properly deals with nested keys."""
        if "." in key:
            key, remainder = key.split(".", 1)
            try:
                del self[key][remainder]
            except KeyError:
                self[key].del_key(remainder)

            # If we removed the last subkey, delete the parent key too
            if len(self[key].keys()) == 0:
                del self[key]

        else:
            del self[key]

    def as_dict(self, flat=False):
        """
        Return the AttrDict as a pure dict (with nested dicts if
        necessary).

        """
        if flat:
            return self.as_dict_flat()
        else:
            return self.as_dict_nested()

    def as_dict_nested(self):
        d = {}
        for k, v in self.items():
            if isinstance(v, AttrDict):
                d[k] = v.as_dict()
            elif isinstance(v, list):
                d[k] = [i if not isinstance(i, AttrDict) else i.as_dict() for i in v]
            else:
                d[k] = v
        return d

    def as_dict_flat(self):
        d = {}
        keys = self.keys_nested()
        for k in keys:
            d[k] = self.get_key(k)
        return d

    def to_yaml(self, path=None):
        """
        Saves the AttrDict to the ``path`` as a YAML file, or returns
        a YAML string if ``path`` is None.

        """
        result = self.copy()
        yaml_ = ruamel_yaml.YAML()
        yaml_.indent = 2
        yaml_.block_seq_indent = 0

        # Numpy objects should be converted to regular Python objects,
        # so that they are properly displayed in the resulting YAML output
        for k in result.keys_nested():
            # Convert numpy numbers to regular python ones
            v = result.get_key(k)
            if isinstance(v, np.floating):
                result.set_key(k, float(v))
            elif isinstance(v, np.integer):
                result.set_key(k, int(v))
            # Lists are turned into seqs so that they are formatted nicely
            elif isinstance(v, list):
                result.set_key(k, yaml_.seq(v))

        result = result.as_dict()

        if path is not None:
            with open(path, "w") as f:
                yaml_.dump(result, f)
        else:
            stream = io.StringIO()
            yaml_.dump(result, stream)
            return stream.getvalue()

    def keys_nested(self, subkeys_as="list"):
        """
        Returns all keys in the AttrDict, sorted, including the keys of
        nested subdicts (which may be either regular dicts or AttrDicts).

        If ``subkeys_as='list'`` (default), then a list of
        all keys is returned, in the form ``['a', 'b.b1', 'b.b2']``.

        If ``subkeys_as='dict'``, a list containing keys and dicts of
        subkeys is returned, in the form ``['a', {'b': ['b1', 'b2']}]``.

        """
        keys = []
        for k, v in sorted(self.items()):
            # Check if dict instance (which AttrDict is too),
            # and for non-emptyness of the dict
            if isinstance(v, dict) and v:
                if subkeys_as == "list":
                    keys.extend([k + "." + kk for kk in v.keys_nested()])
                elif subkeys_as == "dict":
                    keys.append({k: v.keys_nested(subkeys_as=subkeys_as)})
            else:
                keys.append(k)
        return keys

    def union(
        self,
        other,
        allow_override=False,
        allow_replacement=False,
        allow_subdict_override_with_none=False,
    ):
        """
        Merges the AttrDict in-place with the passed ``other``
        AttrDict. Keys in ``other`` take precedence, and nested keys
        are properly handled.

        If ``allow_override`` is False, a KeyError is raised if
        other tries to redefine an already defined key.

        If ``allow_replacement``, allow "_REPLACE_" key to replace an
        entire sub-dict.

        If ``allow_subdict_override_with_none`` is False (default),
        a key of the form ``this.that: None`` in other will be ignored
        if subdicts exist in self like ``this.that.foo: 1``, rather
        than wiping them.

        """
        self_keys = self.keys_nested()
        other_keys = other.keys_nested()
        if allow_replacement:
            WIPE_KEY = "_REPLACE_"
            override_keys = [k for k in other_keys if WIPE_KEY not in k]
            wipe_keys = [
                k.split("." + WIPE_KEY)[0] for k in other_keys if WIPE_KEY in k
            ]
        else:
            override_keys = other_keys
            wipe_keys = []
        for k in override_keys:
            if not allow_override and k in self_keys:
                raise KeyError("Key defined twice: {}".format(k))
            else:
                other_value = other.get_key(k)
                # If other value is None, and would overwrite an entire subdict,
                # we skip it
                if not (
                    other_value is None and isinstance(self.get_key(k, None), AttrDict)
                ):
                    self.set_key(k, other_value)
        for k in wipe_keys:
            self.set_key(k, other.get_key(k + "." + WIPE_KEY))



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

preprocess_checks.py
~~~~~~~~~~~~~~~~~~~~

Checks for model consistency and possible errors during preprocessing.

"""

import os
import logging

import numpy as np
import pandas as pd

from inspect import signature

import calliope
from calliope._version import __version__
from calliope.core.attrdict import AttrDict
from calliope.preprocess.util import get_all_carriers
from calliope.core.util.tools import load_function

logger = logging.getLogger(__name__)

DEFAULTS = AttrDict.from_yaml(
    os.path.join(os.path.dirname(calliope.__file__), "config", "defaults.yaml")
)
POSSIBLE_COSTS = [i for i in DEFAULTS.techs.default_tech.costs.default_cost.keys()]


def check_overrides(config_model, override):
    """
    Perform checks on the override dict and override file inputs to ensure they
    are not doing something silly.
    """
    model_warnings = []
    info = []
    for key in override.as_dict_flat().keys():
        if key in config_model.as_dict_flat().keys():
            info.append(
                "Override applied to {}: {} -> {}".format(
                    key, config_model.get_key(key), override.get_key(key)
                )
            )
        else:
            info.append(
                "`{}`:{} applied from override as new configuration".format(
                    key, override.get_key(key)
                )
            )

    # Check if overriding coordinates are in the same coordinate system. If not,
    # delete all incumbent coordinates, ready for the new coordinates to come in
    if any(["coordinates" in k for k in config_model.as_dict_flat().keys()]) and any(
        ["coordinates" in k for k in override.as_dict_flat().keys()]
    ):

        # get keys that might be deleted and incumbent coordinate system
        config_keys = [
            k for k in config_model.as_dict_flat().keys() if "coordinates." in k
        ]
        config_coordinates = set([k.split("coordinates.")[-1] for k in config_keys])

        # get overriding coordinate system
        override_coordinates = set(
            k.split("coordinates.")[-1]
            for k in override.as_dict_flat().keys()
            if "coordinates." in k
        )

        # compare overriding and incumbent, deleting incumbent if overriding is different
        if config_coordinates != override_coordinates:
            for key in config_keys:
                config_model.del_key(key)
            model_warnings.append(
                "Updated from coordinate system {} to {}, using overrides".format(
                    config_coordinates, override_coordinates
                )
            )

    if info:
        logger.info("\n".join(info))

    return model_warnings


def check_initial(config_model):
    """
    Perform initial checks of model and run config dicts.

    Returns
    -------
    model_warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    errors = []
    model_warnings = []

    # Check for version mismatch
    model_version = config_model.model.get("calliope_version", False)
    if model_version:
        if not str(model_version) in __version__:
            model_warnings.append(
                "Model configuration specifies calliope_version={}, "
                "but you are running {}. Proceed with caution!".format(
                    model_version, __version__
                )
            )

    # Check top-level keys
    for k in config_model.keys():
        if k not in [
            "model",
            "run",
            "locations",
            "tech_groups",
            "techs",
            "links",
            "overrides",
            "scenarios",
            "config_path",
            "group_constraints",
        ]:
            model_warnings.append(
                "Unrecognised top-level configuration item: {}".format(k)
            )

    # Check that all required top-level keys are specified
    for k in ["model", "run", "locations", "techs"]:
        if k not in config_model.keys():
            errors.append(
                "Model is missing required top-level configuration item: {}".format(k)
            )

    # Check run configuration
    # Exclude solver_options and objective_options.cost_class from checks,
    # as we don't know all possible options for all solvers
    for k in config_model["run"].keys_nested():
        if (
            k not in DEFAULTS["run"].keys_nested()
            and "solver_options" not in k
            and "objective_options.cost_class" not in k
        ):
            model_warnings.append(
                "Unrecognised setting in run configuration: {}".format(k)
            )

    # Check model configuration, but top-level keys only
    for k in config_model["model"].keys():
        if k not in DEFAULTS["model"].keys():
            model_warnings.append(
                "Unrecognised setting in model configuration: {}".format(k)
            )
    # If spores run mode is selected, check the correct definition of all needed parameters
    if config_model.run.mode == "spores":
        # Check that spores number is greater than 0, otherwise raise warning
        if config_model.run.spores_options.spores_number == 0:
            model_warnings.append(
                "spores run mode is selected, but a number of 0 spores is requested"
            )
        # Check that slack cost is greater than 0, otherwise set to default (0.1) and raise warning
        if config_model.run.spores_options.slack <= 0:
            config_model.run.spores_options.slack = 0.1
            model_warnings.append(
                "Slack must be greater than zero, setting slack to default value of 0.1 "
            )
        # Check that score_cost_class is a string
        _spores_cost_class = config_model.run.spores_options.get("score_cost_class", {})
        if not isinstance(_spores_cost_class, str):
            errors.append("`run.spores_options.score_cost_class` must be a string")
        # Check that slack_cost_group is one of the defined group contraints
        if (
            config_model.run.spores_options.slack_cost_group
            not in config_model.get("group_constraints", {}).keys()
        ):
            errors.append(
                "`run.spores_options.slack_cost_group` must correspond to "
                "one of the group constraints defined in the model"
            )

    # Only ['in', 'out', 'in_2', 'out_2', 'in_3', 'out_3']
    # are allowed as carrier tiers
    for key in config_model.as_dict_flat().keys():
        if "essentials.carrier_" in key and key.split(".carrier_")[-1].split(".")[
            0
        ] not in ["in", "out", "in_2", "out_2", "in_3", "out_3"]:
            errors.append(
                "Invalid carrier tier found at {}. Only "
                "'carrier_' + ['in', 'out', 'in_2', 'out_2', 'in_3', 'out_3'] "
                "is valid.".format(key)
            )

    # Warn if any unknown group constraints are defined
    permitted_group_constraints = DEFAULTS.group_constraints.default_group.keys()

    for group in config_model.get("group_constraints", {}).keys():
        for key in config_model.group_constraints[group].keys():
            if key not in permitted_group_constraints:
                model_warnings.append(
                    "Unrecognised group constraint `{}` in group `{}` "
                    "will be ignored - possibly a misspelling?".format(key, group)
                )

    # No techs may have the same identifier as a tech_group
    name_overlap = set(config_model.tech_groups.keys()) & set(config_model.techs.keys())
    if name_overlap:
        errors.append(
            "tech_groups and techs with " "the same name exist: {}".format(name_overlap)
        )

    # Checks for techs and tech_groups:
    # * All user-defined tech and tech_groups must specify a parent
    # * techs cannot be parents, only tech groups can
    # * No carrier may be called 'resource'
    default_tech_groups = list(DEFAULTS.tech_groups.keys())
    for tg_name, tg_config in config_model.tech_groups.items():
        if tg_name in default_tech_groups:
            continue
        if not tg_config.get_key("essentials.parent"):
            errors.append(
                "tech_group {} does not define " "`essentials.parent`".format(tg_name)
            )
        elif tg_config.get_key("essentials.parent") in config_model.techs.keys():
            errors.append(
                "tech_group `{}` has a tech as a parent, only another tech_group "
                "is allowed".format(tg_name)
            )
        if "resource" in get_all_carriers(tg_config.essentials):
            errors.append(
                "No carrier called `resource` may "
                "be defined (tech_group: {})".format(tg_name)
            )

    for t_name, t_config in config_model.techs.items():
        for key in t_config.keys():
            if key not in DEFAULTS.techs.default_tech.keys():
                model_warnings.append(
                    "Unknown key `{}` defined for tech {}.".format(key, t_name)
                )
        if not t_config.get_key("essentials.parent"):
            errors.append(
                "tech {} does not define " "`essentials.parent`".format(t_name)
            )
        elif t_config.get_key("essentials.parent") in config_model.techs.keys():
            errors.append(
                "tech `{}` has another tech as a parent, only a tech_group "
                "is allowed".format(tg_name)
            )
        if "resource" in get_all_carriers(t_config.essentials):
            errors.append(
                "No carrier called `resource` may "
                "be defined (tech: {})".format(t_name)
            )

    # Check whether any unrecognised mid-level keys are defined in techs, locations, or links
    for k, v in config_model.get("locations", {}).items():
        unrecognised_keys = [
            i for i in v.keys() if i not in DEFAULTS.locations.default_location.keys()
        ]
        if len(unrecognised_keys) > 0:
            errors.append(
                "Location `{}` contains unrecognised keys {}. "
                "These could be mispellings or a technology not defined "
                "under the `techs` key.".format(k, unrecognised_keys)
            )
        for loc_tech_key, loc_tech_val in v.get("techs", {}).items():
            if loc_tech_val is None:
                continue
            unrecognised_keys = [
                i
                for i in loc_tech_val.keys()
                if i not in DEFAULTS.techs.default_tech.keys()
            ]
            if len(unrecognised_keys) > 0:
                errors.append(
                    "Technology `{}` in location `{}` contains unrecognised keys {}; "
                    "these are most likely mispellings".format(
                        loc_tech_key, k, unrecognised_keys
                    )
                )

    default_link = DEFAULTS.links["default_location_from,default_location_to"]
    for k, v in config_model.get("links", {}).items():
        unrecognised_keys = [i for i in v.keys() if i not in default_link.keys()]
        if len(unrecognised_keys) > 0:
            errors.append(
                "Link `{}` contains unrecognised keys {}. "
                "These could be mispellings or a technology not defined "
                "under the `techs` key.".format(k, unrecognised_keys)
            )
        for link_tech_key, link_tech_val in v.get("techs", {}).items():
            if link_tech_val is None:
                continue
            unrecognised_keys = [
                i
                for i in link_tech_val.keys()
                if i not in default_link.techs.default_tech.keys()
                and i not in DEFAULTS.techs.default_tech.keys()
            ]
            if len(unrecognised_keys) > 0:
                errors.append(
                    "Technology `{}` in link `{}` contains unrecognised keys {}; "
                    "these are most likely mispellings".format(
                        link_tech_key, k, unrecognised_keys
                    )
                )

    # Error if a technology is defined twice, in opposite directions
    link_techs = [
        tuple(sorted(j.strip() for j in k.split(","))) + (i,)
        for k, v in config_model.get("links", {}).items()
        for i in v.get("techs", {}).keys()
    ]
    if len(link_techs) != len(set(link_techs)):
        duplicated_techs = np.array(link_techs)[
            pd.Series(link_techs).duplicated().values
        ]
        duplicated_techs = set([i[-1] for i in duplicated_techs])
        tech_end = "y" if len(duplicated_techs) == 1 else "ies"
        errors.append(
            "Technolog{} {} defined twice on a link defined in both directions "
            "(e.g. `A,B` and `B,A`). A technology can only be defined on one link "
            "even if it allows unidirectional flow in each direction "
            "(i.e. `one_way: true`).".format(tech_end, ", ".join(duplicated_techs))
        )

    # Error if a constraint is loaded from file that must not be
    allowed_from_file = DEFAULTS.model.file_allowed
    for k, v in config_model.as_dict_flat().items():
        if "file=" in str(v):

            possible_identifiers = k.split(".")
            is_time_varying = any("_time_varying" in i for i in possible_identifiers)
            if is_time_varying:
                model_warnings.append(
                    "Using custom constraint `{}` with time-varying data.".format(k)
                )
            elif (
                not set(possible_identifiers).intersection(allowed_from_file)
                and not is_time_varying
            ):
                errors.append(
                    "Cannot load data from file for configuration `{}`.".format(k)
                )

    # We no longer allow cost_class in objective_obtions to be a string
    _cost_class = config_model.run.objective_options.get("cost_class", {})

    if not isinstance(_cost_class, dict):
        errors.append(
            "`run.objective_options.cost_class` must be a dictionary."
            "If you want to minimise or maximise with a single cost class, "
            'use e.g. "{monetary: 1}", which gives the monetary cost class a weight '
            "of 1 in the objective, and ignores any other cost classes."
        )
    else:
        # This next check is only run if we have confirmed that cost_class is
        # a dict, as it errors otherwise

        # For cost minimisation objective, check for cost_class: None and set to one
        for k, v in _cost_class.items():
            if v is None:
                config_model.run.objective_options.cost_class[k] = 1
                model_warnings.append(
                    "cost class {} has weight = None, setting weight to 1".format(k)
                )

    if (
        isinstance(_cost_class, dict)
        and _cost_class.get("monetary", 0) == 1
        and len(_cost_class.keys()) > 1
    ):
        # Warn that {monetary: 1} is still in the objective, since it is not
        # automatically overidden on setting another objective.
        model_warnings.append(
            "Monetary cost class with a weight of 1 is still included "
            "in the objective. If you want to remove the monetary cost class, "
            'add `{"monetary": 0}` to the dictionary nested under '
            " `run.objective_options.cost_class`."
        )

    # Don't allow time clustering with cyclic storage if not also using
    # storage_inter_cluster
    storage_inter_cluster = "model.time.function_options.storage_inter_cluster"
    if (
        config_model.get_key("model.time.function", None) == "apply_clustering"
        and config_model.get_key("run.cyclic_storage", True)
        and not config_model.get_key(storage_inter_cluster, True)
    ):
        errors.append(
            "When time clustering, cannot have cyclic storage constraints if "
            "`storage_inter_cluster` decision variable is not activated."
        )

    return model_warnings, errors


def _check_tech_final(
    model_run, tech_id, tech_config, loc_id, model_warnings, errors, comments
):
    """
    Checks individual tech/tech groups at specific locations.
    NOTE: Updates `model_warnings` and `errors` lists in-place.
    """
    if tech_id not in model_run.techs:
        model_warnings.append(
            "Tech {} was removed by setting ``exists: False`` - not checking "
            "the consistency of its constraints at location {}.".format(tech_id, loc_id)
        )
        return model_warnings, errors

    required = model_run.techs[tech_id].required_constraints
    allowed = model_run.techs[tech_id].allowed_constraints
    allowed_costs = model_run.techs[tech_id].allowed_costs

    # Error if required constraints are not defined
    for r in required:
        # If it's a string, it must be defined
        single_ok = isinstance(r, str) and r in tech_config.constraints
        # If it's a list of strings, one of them must be defined
        multiple_ok = isinstance(r, list) and any(
            [i in tech_config.constraints for i in r]
        )
        if not single_ok and not multiple_ok:
            errors.append(
                "`{}` at `{}` fails to define "
                "all required constraints: {}".format(tech_id, loc_id, required)
            )

    # Warn if defining a carrier ratio for a conversion_plus tech,
    # but applying it to a carrier that isn't one of the carriers specified by that tech
    # e.g. carrier_ratios.carrier_in_2.cooling when cooling isn't a carrier`
    defined_carriers = get_all_carriers(model_run.techs[tech_id].essentials)
    carriers_in_ratios = [
        i.split(".")[-1]
        for i in tech_config.constraints.get_key("carrier_ratios", AttrDict())
        .as_dict_flat()
        .keys()
    ]
    for carrier in carriers_in_ratios:
        if carrier not in defined_carriers:
            model_warnings.append(
                "Tech `{t}` gives a carrier ratio for `{c}`, but does not actually "
                "configure `{c}` as a carrier.".format(t=tech_id, c=carrier)
            )

    # If the technology involves storage, warn when energy_cap and storage_cap aren't connected
    energy_cap_per_storage_cap_params = [
        "charge_rate",
        "energy_cap_per_storage_cap_min",
        "energy_cap_per_storage_cap_max",
        "energy_cap_per_storage_cap_equals",
    ]
    if loc_id + "::" + tech_id in model_run.sets.loc_techs_store and not any(
        i in tech_config.constraints.keys() for i in energy_cap_per_storage_cap_params
    ):
        logger.info(
            "`{}` at `{}` has no constraint to explicitly connect `energy_cap` to "
            "`storage_cap`, consider defining a `energy_cap_per_storage_cap_min/max/equals` "
            "constraint".format(tech_id, loc_id)
        )

    # If a technology is defined by units (i.e. integer decision variable), it must define energy_cap_per_unit
    if (
        any(["units_" in k for k in tech_config.constraints.keys()])
        and "energy_cap_per_unit" not in tech_config.constraints.keys()
    ):
        errors.append(
            "`{}` at `{}` fails to define energy_cap_per_unit when specifying "
            "technology in units_max/min/equals".format(tech_id, loc_id, required)
        )

    # If a technology is defined by units & is a storage tech, it must define storage_cap_per_unit
    if (
        any(["units_" in k for k in tech_config.constraints.keys()])
        and model_run.techs[tech_id].essentials.parent in ["storage", "supply_plus"]
        and any(["storage" in k for k in tech_config.constraints.keys()])
        and "storage_cap_per_unit" not in tech_config.constraints.keys()
    ):
        errors.append(
            "`{}` at `{}` fails to define storage_cap_per_unit when specifying "
            "technology in units_max/min/equals".format(tech_id, loc_id, required)
        )

    # If a technology defines force_resource but is not in loc_techs_finite_resource
    if (
        "force_resource" in tech_config.constraints.keys()
        and loc_id + "::" + tech_id not in model_run.sets.loc_techs_finite_resource
    ):

        model_warnings.append(
            "`{}` at `{}` defines force_resource but not a finite resource, so "
            "force_resource will not be applied".format(tech_id, loc_id)
        )

    # Gather remaining unallowed constraints
    remaining = set(tech_config.constraints) - set(allowed)

    # Error if something is defined that's not allowed, but is in defaults
    # Warn if something is defined that's not allowed, but is not in defaults
    # (it could be a misspelling)
    for k in remaining:
        if k in DEFAULTS.techs.default_tech.constraints.keys():
            errors.append(
                "`{}` at `{}` defines non-allowed "
                "constraint `{}`".format(tech_id, loc_id, k)
            )
        else:
            model_warnings.append(
                "`{}` at `{}` defines unrecognised "
                "constraint `{}` - possibly a misspelling?".format(tech_id, loc_id, k)
            )

    # Error if an `export` statement does not match the given carrier_outs
    if "export_carrier" in tech_config.constraints:
        essentials = model_run.techs[tech_id].essentials
        export = tech_config.constraints.export_carrier
        if export and export not in [
            essentials.get_key(k, "")
            for k in ["carrier_out", "carrier_out_2", "carrier_out_3"]
        ]:
            errors.append(
                "`{}` at `{}` is attempting to export a carrier "
                "not given as an output carrier: `{}`".format(tech_id, loc_id, export)
            )

    # Error if non-allowed costs are defined
    for cost_class in tech_config.get_key("costs", {}):
        for k in tech_config.costs[cost_class]:
            if k not in allowed_costs:
                errors.append(
                    "`{}` at `{}` defines non-allowed "
                    "{} cost: `{}`".format(tech_id, loc_id, cost_class, k)
                )

    # Error if non-allowed `resource_unit` is defined
    if tech_config.constraints.get("resource_unit", "energy") not in [
        "energy",
        "energy_per_cap",
        "energy_per_area",
    ]:
        errors.append(
            "`{}` is an unknown resource unit for `{}` at `{}`. "
            "Only `energy`, `energy_per_cap`, or `energy_per_area` is allowed.".format(
                tech_config.constraints.resource_unit, tech_id, loc_id
            )
        )

    return model_warnings, errors


def check_final(model_run):
    """
    Perform final checks of the completely built model_run.

    Returns
    -------
    comments : AttrDict
        debug output
    model_warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    model_warnings, errors = [], []
    comments = AttrDict()

    # Go through all loc-tech combinations and check validity
    for loc_id, loc_config in model_run.locations.items():
        if "techs" in loc_config:
            for tech_id, tech_config in loc_config.techs.items():
                _check_tech_final(
                    model_run,
                    tech_id,
                    tech_config,
                    loc_id,
                    model_warnings,
                    errors,
                    comments,
                )

        if "links" in loc_config:
            for link_id, link_config in loc_config.links.items():
                for tech_id, tech_config in link_config.techs.items():
                    _check_tech_final(
                        model_run,
                        tech_id,
                        tech_config,
                        "link {}:{}".format(loc_id, link_id),
                        model_warnings,
                        errors,
                        comments,
                    )

    # Either all locations or no locations must have coordinates
    all_locs = list(model_run.locations.keys())
    locs_with_coords = [
        k for k in model_run.locations.keys() if "coordinates" in model_run.locations[k]
    ]
    if len(locs_with_coords) != 0 and len(all_locs) != len(locs_with_coords):
        errors.append(
            "Either all or no locations must have `coordinates` defined. "
            "Locations defined: {} - Locations with coordinates: {}".format(
                all_locs, locs_with_coords
            )
        )

    # If locations have coordinates, they must all be either lat/lon or x/y
    elif len(locs_with_coords) != 0:
        first_loc = list(model_run.locations.keys())[0]
        try:
            coord_keys = sorted(list(model_run.locations[first_loc].coordinates.keys()))
            if coord_keys != ["lat", "lon"] and coord_keys != ["x", "y"]:
                errors.append(
                    "Unidentified coordinate system. All locations must either"
                    "use the format {lat: N, lon: M} or {x: N, y: M}."
                )
        except AttributeError:
            errors.append(
                "Coordinates must be given in the format {lat: N, lon: M} or "
                "{x: N, y: M}, not " + str(model_run.locations[first_loc].coordinates)
            )

        for loc_id, loc_config in model_run.locations.items():
            try:
                if sorted(list(loc_config.coordinates.keys())) != coord_keys:
                    errors.append("All locations must use the same coordinate format.")
                    break
            except AttributeError:
                errors.append(
                    "Coordinates must be given in the format {lat: N, lon: M} or "
                    "{x: N, y: M}, not "
                    + str(model_run.locations[first_loc].coordinates)
                )
                break

    # Ensure that timeseries have no non-unique index values
    for k, df in model_run["timeseries_data"].items():
        if df.index.duplicated().any():
            errors.append(
                "Time series `{}` contains non-unique timestamp values.".format(k)
            )

    # Warn if loc/tech is defined in group constraint that doesn't exist in the model
    for i in ["locs", "techs"]:
        config_i = "locations" if i == "locs" else i
        _missing = set()
        for group, group_vals in model_run.get("group_constraints", {}).items():
            if i in group_vals.keys() and set(group_vals[i]).difference(
                model_run[config_i].keys()
            ):
                _missing.update(
                    set(group_vals[i]).difference(model_run[config_i].keys())
                )
                model_run["group_constraints"][group][i] = list(
                    set(group_vals[i]).intersection(model_run[config_i].keys())
                )
                if model_run["group_constraints"][group][i] == []:
                    model_warnings.append(
                        "Constraint group `{}` will be completely ignored since "
                        "none of the defined {} are valid for this model.".format(
                            group, i
                        )
                    )
                    model_run["group_constraints"][group]["exists"] = False
        if _missing:
            model_warnings.append(
                "Possible misspelling in group constraints: {0} {1} given in "
                "group constraints, but not defined as {0} in the model. They "
                "will be ignored in the optimisation run".format(i, _missing)
            )

    # Warn if objective cost class is not defined elsewhere in the model
    objective_cost_class = set(model_run.run.objective_options.cost_class.keys())
    cost_classes = model_run.sets.costs
    cost_classes_mismatch = objective_cost_class.difference(cost_classes)
    if cost_classes_mismatch:
        model_warnings.append(
            "Cost classes `{}` are defined in the objective options but not "
            "defined elsewhere in the model. They will be ignored in the "
            "objective function.".format(cost_classes_mismatch)
        )

    # FIXME:
    # make sure `comments` is at the the base level:
    # i.e. comments.model_run.xxxxx....

    return comments, model_warnings, errors


def check_model_data(model_data):
    """
    Perform final checks of the completely built xarray Dataset `model_data`.

    Returns
    -------
    comments : AttrDict
        debug output
    model_warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    model_warnings, errors = [], []
    comments = AttrDict()

    # Ensure that no loc-tech specifies infinite resource and force_resource=True
    if "force_resource" in model_data.data_vars:
        relevant_loc_techs = [
            i.loc_techs_finite_resource.item()
            for i in model_data.force_resource
            if i.item() is True
        ]
        forced_resource = model_data.resource.loc[
            dict(loc_techs_finite_resource=relevant_loc_techs)
        ]
        conflict = forced_resource.where(forced_resource == np.inf).to_pandas().dropna()
        if not conflict.empty:
            errors.append(
                "loc_tech(s) {} cannot have `force_resource` set as infinite "
                "resource values are given".format(", ".join(conflict.index))
            )

    # Ensure that if a tech has negative costs, there is a max cap defined
    # FIXME: doesn't consider capacity being set by a linked constraint e.g.
    # `resource_cap_per_energy_cap`.
    relevant_caps = [
        i
        for i in ["energy_cap", "storage_cap", "resource_cap", "resource_area"]
        if "cost_" + i in model_data.data_vars.keys()
    ]
    for cap in relevant_caps:
        relevant_loc_techs = (
            model_data["cost_" + cap]
            .where(model_data["cost_" + cap] < 0, drop=True)
            .to_pandas()
        )
        cap_max = cap + "_max"
        cap_equals = cap + "_equals"
        for loc_tech in relevant_loc_techs.columns:
            try:
                cap_val = model_data[cap_max][loc_tech].item()
            except KeyError:
                try:
                    cap_val = model_data[cap_equals][loc_tech].item()
                except KeyError:
                    cap_val = np.nan
            if np.isinf(cap_val) or np.isnan(cap_val):
                errors.append(
                    "loc_tech {} cannot have a negative cost_{} as the "
                    "corresponding capacity constraint is not set".format(loc_tech, cap)
                )

    for loc_tech in set(model_data.loc_techs_demand.values).intersection(
        model_data.loc_techs_finite_resource.values
    ):
        if any(model_data.resource.loc[loc_tech].values > 0):
            errors.append(
                "Positive resource given for demand loc_tech {}. All demands "
                "must have negative resource".format(loc_tech)
            )

    # Delete all empty dimensions & the variables associated with them
    for dim_name, dim_length in model_data.dims.items():
        if dim_length == 0:
            if dim_name in model_data.coords.keys():
                del model_data[dim_name]
            associated_vars = [
                var
                for var, data in model_data.data_vars.items()
                if dim_name in data.dims
            ]
            model_data = model_data.drop(associated_vars)
            model_warnings.append(
                "dimension {} and associated variables {} were empty, so have "
                "been deleted".format(dim_name, ", ".join(associated_vars))
            )

    # Check if we're allowed to use operate mode
    if "allow_operate_mode" not in model_data.attrs.keys():
        daily_timesteps = [
            model_data.timestep_resolution.loc[i].values
            for i in np.unique(model_data.timesteps.to_index().strftime("%Y-%m-%d"))
        ]
        if not np.all(daily_timesteps == daily_timesteps[0]):
            model_data.attrs["allow_operate_mode"] = 0
            model_warnings.append(
                "Operational mode requires the same timestep resolution profile "
                "to be emulated on each date"
            )
        else:
            model_data.attrs["allow_operate_mode"] = 1

    # Check for any milp constraints, and warn that the problem contains binary /
    # integer decision variables
    if any("_milp_constraint" in i for i in model_data.dims):
        model_warnings.append(
            "Integer and / or binary decision variables are included in this model. "
            "This may adversely affect solution time, particularly if you are "
            "using a non-commercial solver. To improve solution time, consider "
            "changing MILP related solver options (e.g. `mipgap`) or removing "
            "MILP constraints."
        )

    # Check for storage_initial being a fractional value

    if hasattr(model_data, "loc_techs_store"):
        for loc_tech in model_data.loc_techs_store.values:
            if hasattr(model_data, "storage_initial"):
                if (
                    model_data.storage_initial.loc[{"loc_techs_store": loc_tech}].values
                    > 1
                ):
                    errors.append(
                        "storage_initial values larger than 1 are not allowed."
                    )
                if (
                    model_data.storage_initial.loc[{"loc_techs_store": loc_tech}].values
                    == 0
                ):
                    model_data.storage_initial.loc[{"loc_techs_store": loc_tech}] = 0.0

    # Check for storage_initial being greater than or equal to the storage_discharge_depth

    if hasattr(model_data, "loc_techs_store"):
        for loc_tech in model_data.loc_techs_store.values:
            if hasattr(model_data, "storage_initial") and hasattr(
                model_data, "storage_discharge_depth"
            ):
                if (
                    model_data.storage_initial.loc[{"loc_techs_store": loc_tech}].values
                    < model_data.storage_discharge_depth.loc[
                        {"loc_techs_store": loc_tech}
                    ].values
                ):
                    errors.append(
                        "storage_initial is smaller than storage_discharge_depth."
                        " Please change the model configuration to ensure that"
                        " storage initial is greater than or equal to storage_discharge_depth"
                    )

    # Check for storage_inter_cluster not being used together with storage_discharge_depth
    if hasattr(model_data, "clusters") and hasattr(
        model_data, "storage_discharge_depth"
    ):
        errors.append(
            "storage_discharge_depth is currently not allowed when time clustering is active."
        )

    return model_data, comments, model_warnings, errors
